import asyncio
import logging
import os
import socket
import ssl
import sys
import time
import typing
from enum import Enum

from paho.mqtt.client import LOGGING_LEVEL, MQTT_ERR_SUCCESS, Client, connack_string, MQTTv311, MQTTv31, MQTTv5

from paradox.config import config as cfg
from paradox.data.enums import RunState
from paradox.interfaces import ThreadQueueInterface
from paradox.lib import ps

logger = logging.getLogger("PAI").getChild(__name__)


class ConnectionState(Enum):
    NEW = 0
    CONNECTING = 1
    CONNECTED = 2
    DISCONNECTING = 3


RUN_STATE_2_PAYLOAD = {
    RunState.ERROR: "error",
    RunState.INIT: "initializing",
    RunState.PAUSE: "paused",
    RunState.CONNECTED: "connected",
    RunState.RUN: "online",
    RunState.STOP: "stopped",
}

protocol_map = {
    "3.1": MQTTv31,
    "3.1.1": MQTTv311,
    "5": MQTTv5
}

class MQTTConnection():
    client: Client
    _instance = None

    @classmethod
    def get_instance(cls) -> "MQTTConnection":
        if cls._instance is None:
            cls._instance = MQTTConnection()

        return cls._instance

    def __init__(self):
        self.client = Client(
            "pai"+os.urandom(8).hex(),
            protocol=protocol_map.get(str(cfg.MQTT_PROTOCOL), MQTTv311),
            transport=cfg.MQTT_TRANSPORT,
        )
        self._last_pai_status = "unknown"
        self.pai_status_topic = "{}/{}/{}".format(
            cfg.MQTT_BASE_TOPIC, cfg.MQTT_INTERFACE_TOPIC, "pai_status"
        )
        self.availability_topic = "{}/{}/{}".format(
            cfg.MQTT_BASE_TOPIC, cfg.MQTT_INTERFACE_TOPIC, "availability"
        )
        self.client.on_connect = self._on_connect_cb
        self.client.on_disconnect = self._on_disconnect_cb
        self.state = ConnectionState.NEW
        # self.client.enable_logger(logger)

        # self.client.on_subscribe = lambda client, userdata, mid, granted_qos: logger.debug("Subscribed: %s" %(mid))
        # self.client.on_message = lambda client, userdata, message: logger.debug("Message received: %s" % str(message))
        # self.client.on_publish = lambda client, userdata, mid: logger.debug("Message published: %s" % str(mid))

        ps.subscribe(self.on_run_state_change, "run-state")

        self.registrars = []

        if cfg.MQTT_USERNAME is not None and cfg.MQTT_PASSWORD is not None:
            self.client.username_pw_set(username=cfg.MQTT_USERNAME, password=cfg.MQTT_PASSWORD)

        if cfg.MQTT_TLS_CERT_PATH is not None:
            self.client.tls_set(
                ca_certs=cfg.MQTT_TLS_CERT_PATH,
                certfile=None,
                keyfile=None,
                cert_reqs=ssl.CERT_REQUIRED,
                tls_version=ssl.PROTOCOL_TLSv1_2,
                ciphers=None,
            )
            self.client.tls_insecure_set(False)

        self.client.will_set(self.availability_topic, "offline", 0, retain=True)

        self.client.on_log = self.on_client_log

    def on_client_log(self, client, userdata, level, buf):
        level_std = LOGGING_LEVEL[level]
        exc_info = None

        type_, exc, trace = sys.exc_info()
        if exc:  # Can be (socket.error, OSError, WebsocketConnectionError, ...)
            if hasattr(exc, "errno"):
                exc_msg = f"{os.strerror(exc.errno)}({exc.errno})"
                if exc.errno in [22, 49]:
                    level_std = logging.ERROR
                    buf = f"{buf}: Please check MQTT connection settings. Especially MQTT_BIND_ADDRESS and MQTT_BIND_PORT"
            else:
                exc_msg = str(exc)

            buf = f"{buf}: {exc_msg}"
            if "Connection failed" in buf:
                level_std = logging.WARNING

        if level_std > logging.DEBUG:
            logger.log(level_std, buf, exc_info=exc_info)

    def on_run_state_change(self, state: RunState):
        v = RUN_STATE_2_PAYLOAD.get(state, "unknown")
        self._report_pai_status(v)

    def start(self):
        if self.state == ConnectionState.NEW:
            self.client.loop_start()

            # TODO: Some initial connection retry mechanism required
            try:
                self.client.connect_async(
                    host=cfg.MQTT_HOST,
                    port=cfg.MQTT_PORT,
                    keepalive=cfg.MQTT_KEEPALIVE,
                    bind_address=cfg.MQTT_BIND_ADDRESS,
                    bind_port=cfg.MQTT_BIND_PORT,
                )

                self.state = ConnectionState.CONNECTING

                logger.info("MQTT loop started")
            except socket.gaierror:
                logger.exception(
                    "Failed to connect to MQTT (%s:%d)", cfg.MQTT_HOST, cfg.MQTT_PORT
                )

    def stop(self):
        if self.state in [ConnectionState.CONNECTING, ConnectionState.CONNECTED]:
            self.disconnect()
            self.client.loop_stop()
            logger.info("MQTT loop stopped")

    def publish(self, topic, payload=None, *args, **kwargs):
        logger.debug("MQTT: {}={}".format(topic, payload))

        self.client.publish(topic, payload, *args, **kwargs)

    def _call_registars(self, method, *args, **kwargs):
        for r in self.registrars:
            try:
                if hasattr(r, method) and isinstance(
                    getattr(r, method), typing.Callable
                ):
                    getattr(r, method)(*args, **kwargs)
            except:
                logger.exception(
                    'Failed to call "%s" on "%s"', method, r.__class__.__name__
                )

    def register(self, cls):
        self.registrars.append(cls)

        self.start()

    def unregister(self, cls):
        self.registrars.remove(cls)

        if len(self.registrars) == 0:
            self.stop()

    @property
    def connected(self):
        return self.state == ConnectionState.CONNECTED

    def _report_pai_status(self, status):
        self._last_pai_status = status
        self.publish(self.pai_status_topic, status, qos=cfg.MQTT_QOS, retain=True)
        self.publish(
            self.availability_topic,
            "online" if status in ["online", "paused"] else "offline",
            qos=cfg.MQTT_QOS,
            retain=True,
        )

    def _on_connect_cb(self, client, userdata, flags, result, properties=None):
        # called on Thread-6
        if result == MQTT_ERR_SUCCESS:
            logger.info("MQTT Broker Connected")
            self.state = ConnectionState.CONNECTED
            self._report_pai_status(self._last_pai_status)
            self._call_registars("on_connect", client, userdata, flags, result)
        else:
            logger.error(f"Failed to connect to MQTT: {connack_string(result)} ({result})")

    def _on_disconnect_cb(self, userdata, rc, properties=None):
        # called on Thread-6
        if rc == MQTT_ERR_SUCCESS:
            logger.info("MQTT Broker Disconnected")
        else:
            logger.error(f"MQTT Broker unexpectedly disconnected. Code: {rc}")

        self.state = ConnectionState.NEW
        self._call_registars("on_disconnect", self.client, userdata, rc)

    def disconnect(self, reasoncode=None, properties=None):
        self.state = ConnectionState.DISCONNECTING
        self._report_pai_status("offline")
        self.client.disconnect()

    def message_callback_add(self, *args, **kwargs):
        self.client.message_callback_add(*args, **kwargs)

    def subscribe(self, *args, **kwargs):
        self.client.subscribe(*args, **kwargs)


class AbstractMQTTInterface(ThreadQueueInterface):
    """Interface Class using MQTT"""

    def __init__(self, alarm):
        super().__init__(alarm)

        self.mqtt = MQTTConnection.get_instance()
        self.republish_cache = {}

    def start(self):
        super().start()
        self.mqtt.register(self)
        logger.debug("Registars: %d", len(self.mqtt.registrars))

    def stop(self):
        """ Stops the MQTT Interface Thread"""

        def stop_loop():
            self.republish_task.cancel()
            self.loop.stop()

        self.loop.call_soon_threadsafe(stop_loop)

        self.mqtt.unregister(self)

        super().stop()

    async def republish_loop(self):
        while True:
            await asyncio.sleep(cfg.MQTT_REPUBLISH_INTERVAL)
            trigger = time.time() - cfg.MQTT_REPUBLISH_INTERVAL

            for k, v in filter(
                lambda f: f[1].get("last_publish") <= trigger,
                self.republish_cache.items(),
            ):
                self.publish(k, v["value"], v["qos"], v["retain"])

    def _run(self):
        super(AbstractMQTTInterface, self)._run()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.republish_task = self.loop.create_task(self.republish_loop())

        self.loop.run_forever()
        self.loop.close()

    def publish(self, topic, value, qos, retain):
        self.republish_cache[topic] = {
            "value": value,
            "qos": qos,
            "retain": retain,
            "last_publish": time.time(),
        }
        self.loop.call_soon_threadsafe(self.mqtt.publish, topic, value, qos, retain)

    def _publish_command_status(self, message):
        if cfg.MQTT_PUBLISH_COMMAND_STATUS:
            self.publish(
                f"{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_INTERFACE_TOPIC}/{cfg.MQTT_COMMAND_STATUS_TOPIC}",
                message,
                qos=2,
                retain=True,
            )

    def subscribe_callback(self, sub, callback: typing.Callable):
        self.mqtt.message_callback_add(sub, callback)
        self.mqtt.subscribe(sub)

    def on_disconnect(self, client, userdata, rc):
        """ Called from MQTT connection """
        pass

    def on_connect(self, client, userdata, flags, result):
        """ Called from MQTT connection """
        pass



import logging
import os
import re
import sys


class Config:
    DEFAULTS = {
        "LOGGING_LEVEL_CONSOLE": logging.INFO,  # See documentation of Logging package
        "LOGGING_LEVEL_FILE": logging.ERROR,
        "LOGGING_FILE": (
            None,
            [type(None), str],
        ),  # or set to file path : '/var/log/paradox.log'
        "LOGGING_FILE_MAX_SIZE": (10, int, (0, 0xFFFFFFFF)),  # Max log file size in MB
        "LOGGING_FILE_MAX_FILES": (
            2,
            int,
            (0, 0xFFFFFFFF),
        ),  # Max old log files to keep
        "LOGGING_DUMP_PACKETS": False,  # Dump RAW Packets to log
        "LOGGING_DUMP_MESSAGES": False,  # Dump Messages to log
        "LOGGING_DUMP_STATUS": False,  # Dump Status to log
        "LOGGING_DUMP_EVENTS": False,  # Dump Event details to log
        # Development
        "DEVELOPMENT_DUMP_MEMORY": False,
        # Connection Type
        "CONNECTION_TYPE": ("Serial", str, ["IP", "Serial"]),  # Serial or IP
        # Serial Connection Details
        "SERIAL_PORT": "/dev/ttyS1",  # Pathname of the Serial Port
        "SERIAL_BAUD": 9600,  # Baud rate of the Serial Port. Use 38400(default setting) or 57600 for EVO
        # IP Connection Details
        "IP_CONNECTION_HOST": "127.0.0.1",  # IP Module address when using direct IP Connection
        "IP_CONNECTION_PORT": (
            10000,
            int,
            (1, 65535),
        ),  # IP Module port when using direct IP Connection
        "IP_CONNECTION_PASSWORD": (
            "paradox",
            [str, bytes, type(None)],
        ),  # IP Module password. "paradox" is default.
        "IP_CONNECTION_SITEID": (
            None,
            [str, type(None)],
        ),  # If defined, connection will be made through this method.
        "IP_CONNECTION_EMAIL": (
            None,
            [str, type(None)],
        ),  # Email registered in the site
        "IP_CONNECTION_PANEL_SERIAL": (
            None,
            [str, type(None)],
        ),  # Serial number to be used in multi-panel sites. None for first
        "IP_CONNECTION_BARE": False,  # IP endpoint connects directly to panel. Used for Serial Tunnels over TCP
        # Paradox
        "KEEP_ALIVE_INTERVAL": 10,  # Interval between status updates
        "IO_TIMEOUT": 0.5,  # Timeout for IO operations
        "LIMITS": {},  # By default all zones will be monitored
        "LABEL_ENCODING": "paradox-en",  # Encoding to use when decoding labels. paradox-* or https://docs.python.org/3/library/codecs.html#standard-encodings
        "LABEL_REFRESH_INTERVAL": (
            15 * 60,
            int,
            (0, 0xFFFFFFFF),
        ),  # Interval between refresh of labels
        "OUTPUT_PULSE_DURATION": (
            1,
            float,
            (0, 0xFFFFFFFF),
        ),  # Duration of a PGM pulse in seconds
        "SYNC_TIME": False,  # Update panel time periodically when time drifts more than SYNC_TIME_MIN_DRIFT
        "SYNC_TIME_MIN_DRIFT": (
            120,
            int,
            (120, 0xFFFFFFFF),
        ),  # Minimum time drift in seconds to initiate time sync
        "SYNC_TIME_TIMEZONE": "",  # By default pai uses the same timezone as pai host
        "PASSWORD": (
            None,
            [int, str, bytes, type(None)],
        ),  # PC Password. Set to None if Panel has no Password. In Babyware: Right click on your panel -> Properties -> PC Communication (Babyware) -> PC Communication (Babyware) Tab.
        "PUSH_UPDATE_WITHOUT_CHANGE": False,  # Always notify interfaces of all changes
        # MQTT
        "MQTT_ENABLE": False,  # Enable MQTT Interface
        "MQTT_HOST": "127.0.0.1",  # Hostname or address
        "MQTT_PORT": (
            1883,
            int,
            (1, 65535),
        ),  # TCP Port (TLS port if MQTT_TLS_CERT_PATH is set)
        "MQTT_TLS_CERT_PATH": (
            None,
            [str, type(None)],
        ),  # Path to ca cert (/etc/pai/certs/ca.pem), if you want TLS
        "MQTT_KEEPALIVE": (60, int, (1, 3600)),  # Keep alive
        "MQTT_USERNAME": (None, [str, type(None)]),  # MQTT Username for authentication
        "MQTT_PASSWORD": (None, [str, type(None)]),  # MQTT Password
        "MQTT_RETAIN": True,  # Publish messages with Retain
        "MQTT_QOS": (
            0,
            int,
            (0, 2),
        ),  # Publish messages with QOS level (0 - fire and forget, 1 - at least once, 2 - exactly once)
        "MQTT_PROTOCOL": ("3.1.1", str, ("3.1", "3.1.1", "5")),  # Protocol to use
        "MQTT_TRANSPORT": ("tcp", str, ("tcp", "websockets")),  # Transport to use
        "MQTT_BIND_ADDRESS": "",  # MQTT Bind address (Paho default)
        "MQTT_BIND_PORT": 0,  # MQTT Bind port (Paho default)
        "MQTT_REPUBLISH_INTERVAL": (
            60 * 60 * 12,
            int,
            (60, 0xFFFFFFFF),
        ),  # Interval for republishing all data
        "MQTT_HOMEASSISTANT_AUTODISCOVERY_ENABLE": False,
        "MQTT_HOMEASSISTANT_CODE": (None, [str, type(None)]),
        "MQTT_HOMEASSISTANT_ENTITY_PREFIX": "",  # If you want to prefix all entities you can use: "Paradox {serial_number} ", placeholders "serial_number" and "model" are supported. Default: "" - no prefix
        # MQTT Topics
        "MQTT_BASE_TOPIC": "paradox",  # Root of all topics
        "MQTT_ZONE_TOPIC": "zones",  # Base for zone states
        "MQTT_PARTITION_TOPIC": "partitions",  # Base for partition states
        "MQTT_BUS_TOPIC": "buses",  # Base for buses states
        "MQTT_MODULE_TOPIC": "bus-module",  # Base for bus module states
        "MQTT_SYSTEM_TOPIC": "system",  # Base for panel states
        "MQTT_REPEATER_TOPIC": "repeaters",  # Base for repeater states
        "MQTT_USER_TOPIC": "users",  # Base for user states
        "MQTT_EVENTS_TOPIC": "events",  # Base for events
        "MQTT_CONTROL_TOPIC": "control",  # Base for control of other elements (ROOT/CONTROL/TYPE)
        "MQTT_DEFINITION_TOPIC": "definitions",  # Base for definitions
        "MQTT_HOMEASSISTANT_CONTROL_TOPIC": "hass_control",
        "MQTT_HOMEASSISTANT_DISCOVERY_PREFIX": "homeassistant",
        "MQTT_OUTPUT_TOPIC": "outputs",
        "MQTT_DOOR_TOPIC": "doors",
        "MQTT_KEYPAD_TOPIC": "keypads",
        "MQTT_STATES_TOPIC": "states",
        "MQTT_RAW_TOPIC": "raw",
        "MQTT_NOTIFICATIONS_TOPIC": "notifications",
        "MQTT_SEND_PANIC_TOPIC": "panic",
        "MQTT_PUBLISH_RAW_EVENTS": True,
        "MQTT_PUBLISH_DEFINITIONS": False,  # Publish definitions of partitions/zones/users to mqtt.
        "MQTT_PREFIX_DEVICE_NAME": False,  # Add device ID as prefix to entity names: Paradox 12345678
        "MQTT_INTERFACE_TOPIC": "interface",
        "MQTT_TOGGLE_CODES": {},
        "MQTT_USE_NUMERIC_STATES": False,  # use 0 and 1 instead of True and False
        "MQTT_DASH_PUBLISH": False,
        "MQTT_DASH_TOPIC": "metrics/exchange/pai",
        "MQTT_DASH_TEMPLATE": "/etc/pai/mqtt_dash.txt",
        "MQTT_CHALLENGE_SECRET": (
            None,
            [str, type(None)],
        ),  # MQTT Command authorization challenge
        "MQTT_CHALLENGE_TOPIC": "challenge",
        "MQTT_CHALLENGE_ROUNDS": 1000,
        # Interfaces text command alias
        "COMMAND_ALIAS": {  # alias for commands through text based interfaces
            "arm": "partition all arm",
            "disarm": "partition all disarm",
        },
        "MQTT_PUBLISH_COMMAND_STATUS": False,
        "MQTT_COMMAND_STATUS_TOPIC": "command_status",
        # MQTT command aliases
        "MQTT_COMMAND_ALIAS": {
            # For homebridge
            "armed_home": "arm_stay",
            "armed_night": "arm_sleep",
            "armed_away": "arm",
            "disarmed": "disarm",
        },
        "HOMEASSISTANT_PUBLISH_PARTITION_PROPERTIES": [  # List of partition properties to publish
            "target_state",
            "current_state",
        ],
        "HOMEASSISTANT_PUBLISH_ZONE_PROPERTIES": [  # List of zone properties to publish
            "open",
            "tamper",
        ],
        # Home Assistant Notifications (HASS.io required)
        "HOMEASSISTANT_NOTIFICATIONS_ENABLE": False,
        "HOMEASSISTANT_NOTIFICATIONS_API_URL": "http://supervisor/core/api/services/:domain/:service",
        "HOMEASSISTANT_NOTIFICATIONS_API_TOKEN": "",  # Authentication token used for Home Assistant if not using Supervisor
        "HOMEASSISTANT_NOTIFICATIONS_LOVELACE_URI": "",  # URI to open when notification is clicked
        "HOMEASSISTANT_NOTIFICATIONS_NOTIFIER_NAME": "notify",
        "HOMEASSISTANT_NOTIFICATIONS_MIN_EVENT_LEVEL": (
            "INFO",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
        "HOMEASSISTANT_NOTIFICATIONS_IGNORE_EVENTS": [],  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "HOMEASSISTANT_NOTIFICATIONS_ALLOW_EVENTS": [],  # Same as before but as a white list. Default is use EVENT_FILTERS
        "HOMEASSISTANT_NOTIFICATIONS_EVENT_FILTERS": [  # list of tags, property changes to include or exclude. See event.py for tag list
            "live,alarm,-restore",
            "trouble,-clock",
            "live,tamper",
        ],
        # Pushbullet
        "PUSHBULLET_ENABLE": False,
        "PUSHBULLET_KEY": "",  # Authentication key used for Pushbullet
        "PUSHBULLET_DEVICE": "pai",  # Destination device for notifications
        "PUSHBULLET_CONTACTS": [],  # Pushbullet user identifiers for notifications and interaction
        "PUSHBULLET_IGNORE_EVENTS": [],  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "PUSHBULLET_ALLOW_EVENTS": [],  # Same as before but as a white list. Default is use EVENT_FILTERS
        "PUSHBULLET_EVENT_FILTERS": [  # list of tags, property changes to include or exclude. See event.py for tag list
            "live,alarm,-restore",
            "trouble,-clock",
            "live,tamper",
        ],
        "PUSHBULLET_MIN_EVENT_LEVEL": (
            "INFO",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
        # Pushover
        "PUSHOVER_ENABLE": False,
        "PUSHOVER_KEY": "",  # Application token for Pushover
        "PUSHOVER_BROADCAST_KEYS": [  # Pushover user or group keys to broadcast notifications to
            #    {'user_key':'<user_key>', 'devices': '*'}                # value can be '*' or comma separated list of device names
        ],
        "PUSHOVER_IGNORE_EVENTS": [],  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "PUSHOVER_ALLOW_EVENTS": [],  # Same as before but as a white list. Default is use EVENT_FILTERS
        "PUSHOVER_EVENT_FILTERS": [  # list of tags, property changes to include or exclude. See event.py for tag list
            "live,alarm,-restore",
            "trouble,-clock",
            "live,tamper",
        ],
        "PUSHOVER_MIN_EVENT_LEVEL": (
            "INFO",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
        # Signal
        "SIGNAL_ENABLE": False,
        "SIGNAL_CONTACTS": [],  # Contacts that are allowed to control the panel and receive notifications through Signal
        "SIGNAL_IGNORE_EVENTS": [],  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "SIGNAL_ALLOW_EVENTS": [],  # Same as before but as a white list. Default is use EVENT_FILTERS
        "SIGNAL_EVENT_FILTERS": [  # list of tags, property changes to include or exclude. See event.py for tag list
            "live,alarm,-restore",
            "trouble,-clock",
            "live,tamper",
        ],
        "SIGNAL_MIN_EVENT_LEVEL": (
            "INFO",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
        # GSM
        "GSM_ENABLE": False,
        "GSM_MODEM_BAUDRATE": (
            115200,
            int,
            (9600, 115200),
        ),  # Baudrate of the GSM modem
        "GSM_MODEM_PORT": "",  # Pathname of the serial port
        "GSM_CONTACTS": [],  # Contacts that are allowed to control the panel and receive notifications through SMS
        "GSM_IGNORE_EVENTS": [],  # List of tuples or regexp matching "type,label,property=value,property2=value" eg. [(major, minor), "zone:HOME:entry_delay=True", ...]
        "GSM_ALLOW_EVENTS": [],  # Same as before but as a white list. Default is use EVENT_FILTERS
        "GSM_EVENT_FILTERS": [  # list of tags, property changes to include or exclude. See event.py for tag list
            "live,zone,alarm,trigger"
        ],
        "GSM_MIN_EVENT_LEVEL": (
            "CRITICAL",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
        # IP Socket Interface
        "IP_INTERFACE_ENABLE": False,
        "IP_INTERFACE_BIND_ADDRESS": "0.0.0.0",
        "IP_INTERFACE_BIND_PORT": (10000, int, (1, 65535)),
        "IP_INTERFACE_PASSWORD": ("paradox", [str, bytes, type(None)]),
        # Dummy Interface for testing
        "DUMMY_INTERFACE_ENABLE": False,
        "DUMMY_EVENT_FILTERS": [
            "live,alarm-restore",
            "live,trouble-clock",
            "live,tamper",
            "live,arm",
            "live,disarm",
        ],
        "DUMMY_MIN_EVENT_LEVEL": (
            "DEBUG",
            str,
            ["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        ),
    }

    CONFIG_LOADED = False
    CONFIG_FILE_LOCATION = (None,)

    def __dir__(self):
        return list(self.DEFAULTS.keys()) + list(self.__class__.__dict__) + dir(super())

    def __init__(self):
        self._reset_defaults()

    def load(self, alt_location=None):
        self.CONFIG_LOADED = False

        self._find_config(alt_location)

        sys.stdout.write(
            "Attempting to load configuration from %s\n" % self.CONFIG_FILE_LOCATION
        )
        entries = self._read_config()
        self._update_from_environment(entries)

        # Set values
        for k, v in entries.items():
            if k[0].isupper() and k in self.DEFAULTS:
                default = self.DEFAULTS.get(k)

                if isinstance(default, tuple) and 2 <= len(default) <= 3:
                    default_type = default[1]

                    if not isinstance(default_type, list):
                        default_type = [default_type]

                else:
                    default_type = [type(default)]

                if float in default_type and int not in default_type:
                    default_type.append(int)
                if int in default_type and float not in default_type:
                    default_type.append(float)
                if type(v) in default_type:
                    setattr(self, k, v)
                else:
                    err = "Error parsing configuration: Invalid value type {} for config argument {}. Allowed are: {}".format(
                        type(v), k, default_type
                    )
                    sys.stderr.write(err + "\n")
                    raise (Exception(err))

                if isinstance(default, tuple) and len(default) == 3:
                    expected_value = default[2]
                    valid = False

                    if isinstance(v, int):
                        if expected_value[0] <= v <= expected_value[1]:
                            valid = True
                    elif isinstance(v, str):
                        if v in expected_value:
                            valid = True
                    else:
                        valid = True

                    if valid:
                        setattr(self, k, v)
                    else:
                        err = "Error parsing configuration value: Invalid value for config argument {} (type {}). Allowed are in: {}".format(
                            k, type(v), expected_value
                        )
                        sys.stderr.write(err + "\n")
                        raise (Exception(err))

        self.CONFIG_LOADED = True

    def _update_from_environment(self, entries):
        # Updates values from env variables
        for args in os.environ:
            if not args.startswith("PAI_") or len(args) < 5:
                continue
            opt = args[4:]
            if opt in self.DEFAULTS:
                v = os.environ[args]
                if v.isdigit():
                    v = int(v)

                entries[opt] = v

    def _read_config(self):
        entries = {}
        conf_extension = os.path.splitext(self.CONFIG_FILE_LOCATION)[1]
        if conf_extension in [".conf", ".py"]:
            with open(self.CONFIG_FILE_LOCATION) as f:
                exec(f.read(), None, entries)
        elif conf_extension in [".json"]:
            import json

            with open(self.CONFIG_FILE_LOCATION) as f:
                entries = json.load(f)
        elif conf_extension in [".yaml"]:
            import yaml

            with open(self.CONFIG_FILE_LOCATION) as f:
                entries = yaml.safe_load(f)
        else:
            err = "ERROR: Unsupported configuration file type"
            sys.stderr.write(err + "\n")
            raise (Exception(err))
        return entries

    def _find_config(self, alt_location):
        env_config_path = os.environ.get("PAI_CONFIG_FILE")
        if alt_location is not None:
            locations = [alt_location]
        elif env_config_path:
            locations = [env_config_path]
        else:
            filenames = ["pai.conf", "pai.json", "pai.yaml"]
            locations = [
                os.path.join(dir, filename)
                for dir in [
                    os.path.realpath(os.getcwd()),
                    os.path.expanduser("~/.local/etc"),
                    "/etc/pai",
                    "/usr/local/etc/pai",
                ]
                for filename in filenames
            ]
        for location in locations:
            location = os.path.expanduser(location)
            if os.path.exists(location) and os.path.isfile(location):
                self.CONFIG_FILE_LOCATION = location
                break
        else:
            err = f"ERROR: Could not find configuration file. Tried: {locations}"
            sys.stderr.write(err + "\n")
            raise (Exception(err))

    def _reset_defaults(self):
        # Reset defaults
        for k, v in self.DEFAULTS.items():
            if isinstance(v, tuple):
                v = v[0]

            setattr(self, k, v)


config = Config()

comma_splitter_re = re.compile(r"\s*,\s*")
range_re = re.compile(r"(\d+)\s*-(\d+)\s*$")


def string_to_id_list(input: str):
    arr = []
    blocks = comma_splitter_re.split(input)
    for block in blocks:
        range_match = range_re.match(block)
        if range_match:
            arr += range(int(range_match.group(1)), int(range_match.group(2)) + 1)
        else:
            try:
                arr.append(int(block))
            except ValueError:
                pass

    return arr


def get_limits_for_type(elem_type: str, default: list = None):
    limits = config.LIMITS.get(elem_type, default)
    if limits is None:
        return default

    if isinstance(limits, list):
        return limits

    if isinstance(limits, range):
        return list(limits)

    if isinstance(limits, str):
        if "auto" in limits:
            return default
        return string_to_id_list(limits)



from abc import ABC, abstractmethod
import asyncio
import logging

from construct import Container

from paradox.config import config as cfg
from paradox.connections.connection import Connection
from paradox.connections.handler import IPConnectionHandler
from paradox.connections.ip.commands import IPModuleConnectCommand
from paradox.connections.ip.stun_session import StunSession
from paradox.connections.protocols import IPConnectionProtocol, SerialConnectionProtocol
from paradox.exceptions import PAICriticalException
from paradox.lib.handlers import FutureHandler, HandlerRegistry

logger = logging.getLogger("PAI").getChild(__name__)


class MultiAttemptConnection(Connection):
    async def connect(self) -> bool:
        tries = 1
        max_tries = 3
        self.connected = False

        while tries <= max_tries:
            logger.info("Connecting. Try %d/%d" % (tries, max_tries))
            try:
                await self._try_connect()
                return True
            except asyncio.TimeoutError as e:
                logger.error(
                    "Timeout while establishing connection (try %d/%d): %s"
                    % (tries, max_tries, str(e))
                )
            except OSError as e:
                logger.error(
                    "Connect failed (try %d/%d): %s" % (tries, max_tries, str(e))
                )
            except PAICriticalException:
                raise
            except Exception as e:
                logger.exception(
                    "Unhandled exception while connecting (try %d/%d): %s"
                    % (tries, max_tries, str(e))
                )

            tries += 1

        return False

    @abstractmethod
    async def _try_connect(self):
        raise NotImplementedError("Implement in a subclass")


class BareIPConnection(MultiAttemptConnection):
    def __init__(self, host="127.0.0.1", port=10000):
        super().__init__()
        self.host = host
        self.port = port

    async def _try_connect(self):
        _, self._protocol = await asyncio.get_event_loop().create_connection(
            self._make_protocol, host=self.host, port=self.port
        )

        self.connected = True

    def _make_protocol(self):
        return SerialConnectionProtocol(self)


class IPConnectionWithEncryption(MultiAttemptConnection, IPConnectionHandler, ABC):
    _protocol: IPConnectionProtocol

    def __init__(self, password=None):
        super().__init__()
        if isinstance(password, str):
            password = password.encode()
        self.password = password
        self.key = password

        self.ip_handler_registry = HandlerRegistry()

    def reset_key(self):
        self.set_key(self.password)

    def set_key(self, value):
        self.key = value
        self._protocol.key = value

    def on_ip_message(self, container: Container):
        return asyncio.get_event_loop().create_task(
            self.ip_handler_registry.handle(container)
        )

    async def wait_for_ip_message(self, timeout=cfg.IO_TIMEOUT) -> Container:
        future = FutureHandler()
        return await self.ip_handler_registry.wait_until_complete(future, timeout)

    async def send_raw_ip_message(self, msg):
        self._protocol.send_raw(msg)

    def _make_protocol(self):
        return IPConnectionProtocol(self, self.key)


class LocalIPConnection(IPConnectionWithEncryption):
    def __init__(
        self,
        host,
        port,
        password,
    ):
        super().__init__(password)
        self.host = host
        self.port = port

    async def _try_connect(self) -> None:
        _, self._protocol = await asyncio.get_event_loop().create_connection(
            self._make_protocol, host=self.host, port=self.port
        )

        await IPModuleConnectCommand(self).execute()

        self.connected = True


class StunIPConnection(IPConnectionWithEncryption):
    def __init__(self, site_id, email, panel_serial, password):
        super().__init__(password)

        self.stun_session = StunSession(site_id, email, panel_serial)

    def on_connection_loss(self):
        super().on_connection_loss()

        if self.stun_session is not None:
            self.stun_session.close()

    def write(self, data: bytes):
        """Write data to socket"""

        if self.stun_session is not None:
            self.stun_session.refresh_session_if_required()

        return super().write(data)

    async def _try_connect(self) -> None:
        await self.stun_session.connect()
        _, self._protocol = await asyncio.get_event_loop().create_connection(
            self._make_protocol, sock=self.stun_session.get_socket()
        )

        await IPModuleConnectCommand(self).execute()

        self.connected = True



import asyncio
import json
import logging
from collections import namedtuple

from paradox.config import config as cfg
from paradox.lib import ps
from paradox.lib.utils import SerializableToJSONEncoder
from .entities.abstract_entity import AbstractEntity
from .entities.device import Device
from .entities.factory import MQTTAutodiscoveryEntityFactory

from ...data.model import DetectedPanel
from .core import AbstractMQTTInterface

logger = logging.getLogger("PAI").getChild(__name__)

PreparseResponse = namedtuple("preparse_response", "topics element content")


class HomeAssistantMQTTInterface(AbstractMQTTInterface):
    def __init__(self, alarm):
        super().__init__(alarm)
        self.armed = dict()
        self.partitions = {}
        self.zones = {}
        self.pgms = {}

        self.entity_factory = MQTTAutodiscoveryEntityFactory(self.mqtt.availability_topic)

        self.run_status_topic = self.mqtt.pai_status_topic

        self.connected_future = (
            asyncio.Future()
        )  # TODO: do not create it, use some other
        panel_detected_future = asyncio.Future()
        first_status_update_future = asyncio.Future()

        def _ready_future_callback(x):
            self._publish_when_ready(
                panel_detected_future.result()["panel"],
                first_status_update_future.result()["status"],
            )

        ready_future = asyncio.ensure_future(
            asyncio.gather(
                self.connected_future, panel_detected_future, first_status_update_future
            )
        )
        ready_future.add_done_callback(_ready_future_callback)

        ps.subscribe(panel_detected_future, "panel_detected")
        ps.subscribe(self._handle_labels_loaded, "labels_loaded")
        ps.subscribe(first_status_update_future, "status_update")

    def on_connect(self, client, userdata, flags, result):
        # TODO: do not create connected_future, use some other
        if not self.connected_future.done():
            self.connected_future.set_result(True)

    def _handle_labels_loaded(self, data):
        partitions = data.get("partition", {})
        for k, v in partitions.items():
            p_data = {}
            p_data.update(v)
            self.partitions[k] = p_data

        self.zones = data.get("zone", {})
        self.pgms = data.get("pgm", {})

    def _publish_when_ready(self, panel: DetectedPanel, status):
        self.entity_factory.set_device(Device(panel))

        self._publish_pai_state_sensor_config()
        if "partition" in status:
            self._publish_partition_configs(status["partition"])
        if "zone" in status:
            self._publish_zone_configs(status["zone"])
        if "pgm" in status:
            self._publish_pgm_configs(status["pgm"])
        if "system" in status:
            self._publish_system_property_configs(status['system'])

    def _publish_config(self, entity: AbstractEntity):
        self.publish(entity.configuration_topic, json.dumps(entity, cls=SerializableToJSONEncoder), 0,
                     cfg.MQTT_RETAIN)

    def _publish_pai_state_sensor_config(self):
        pai_state_sensor_config = self.entity_factory.make_pai_status_sensor(self.run_status_topic)
        self._publish_config(pai_state_sensor_config)

    def _publish_partition_configs(self, partition_statuses):
        for partition_key, partition_status in partition_statuses.items():
            if partition_key not in self.partitions:
                continue

            partition = self.partitions[partition_key]
            code = cfg.MQTT_HOMEASSISTANT_CODE or None  # returns None on empty string. For HA Addon Schema parsing

            partition_alarm_control_panel_config = self.entity_factory.make_alarm_control_panel_config(partition, code)
            self._publish_config(partition_alarm_control_panel_config)
            
            # Publish individual entities
            for property_name in partition_status:
                if property_name not in cfg.HOMEASSISTANT_PUBLISH_PARTITION_PROPERTIES:
                    continue
                partition_property_binary_sensor_config = self.entity_factory.make_partition_status_binary_sensor(partition, property_name)
                self._publish_config(partition_property_binary_sensor_config)

    def _publish_zone_configs(self, zone_statuses):
        for zone_key, zone_status in zone_statuses.items():
            if zone_key not in self.zones:
                continue

            zone = self.zones[zone_key]

            # Publish Status
            for property_name in zone_status:
                if property_name not in cfg.HOMEASSISTANT_PUBLISH_ZONE_PROPERTIES:
                    continue
                if property_name == "bypassed":
                    zone_status_sensor = self.entity_factory.make_zone_bypass_switch(zone)
                elif property_name == "signal_strength":
                    zone_status_sensor = self.entity_factory.make_zone_status_numeric_sensor(zone, property_name)
                else:
                    zone_status_sensor = self.entity_factory.make_zone_status_binary_sensor(zone, property_name)
                self._publish_config(zone_status_sensor)

    def _publish_pgm_configs(self, pgm_statuses):
        for pgm_key, pgm_status in pgm_statuses.items():
            if pgm_key not in self.pgms:
                continue

            pgm = self.pgms[pgm_key]

            pgm_switch_config = self.entity_factory.make_pgm_switch(pgm)
            self._publish_config(pgm_switch_config)
    
    def _publish_system_property_configs(self, system_statuses):
        for system_key, system_status in system_statuses.items():
            for property_name in system_status:
                system_property_config = self.entity_factory.make_system_status(system_key, property_name)
                self._publish_config(system_property_config)



from paradox.interfaces.mqtt.entities.abstract_entity import AbstractEntity
from paradox.lib.utils import sanitize_key


class PAIStatusSensor(AbstractEntity):
    def __init__(self, run_status_topic, device, availability_topic):
        super().__init__(device, availability_topic)

        self.hass_entity_type = "sensor"

        self.run_status_topic = run_status_topic

    def serialize(self):
        config = super().serialize()

        del config["availability_topic"]

        return config

    @property
    def state_topic(self):
        return self.run_status_topic

    @property
    def entity_id(self):
        return "pai_status"

    @property
    def entity_name(self):
        return "PAI Status"


class ZoneNumericSensor(AbstractEntity):
    def __init__(self, entity, property, device, availability_topic):
        super().__init__(device, availability_topic)

        self.property = property
        self.label = entity.get("label", entity["key"].replace("_", " "))

        self.key = sanitize_key(entity["key"])

        self.pai_entity_type = "zone"
        self.hass_entity_type = "sensor"


class SystemStatusSensor(AbstractEntity):
    def __init__(self, key, property, device, availability_topic):
        super().__init__(device, availability_topic)

        self.property = property
        self.label = key.replace("_", " ").title()

        self.key = sanitize_key(key)

        self.pai_entity_type = "system"
        self.hass_entity_type = "sensor"

    def serialize(self):
        config = super().serialize()
        if self.key == 'power':
            config['unit_of_measurement'] = 'V'

        return config


import logging
import re
from collections import namedtuple

from paradox.event import ChangeEvent, Event, EventLevel, LiveEvent

re_match = re.compile(
    r"[+-]?(?P<quote>['\"])?(?(quote).+?|[a-z_A-Z=]+)(?(quote)(?P=quote))"
)
re_unquote = re.compile(r"^(?P<quote>['\"])(.*)(?P=quote)$")

logger = logging.getLogger("PAI").getChild(__name__)


def _unquote(s: str):
    return re_unquote.sub(r"\2", s)


class EventFilter:
    def __init__(self, min_level=EventLevel.INFO):
        assert isinstance(min_level, EventLevel)
        self.min_level = min_level

    def match(self, event: Event):
        return event.level >= self.min_level


class LiveEventFilter(EventFilter):
    def match(self, event: Event):
        return super().match(event) and isinstance(event, LiveEvent)


TagQuery = namedtuple(
    "TagQuery", ["include", "exclude", "changes_include", "changes_exclude"]
)
KeyValue = namedtuple("KeyValue", ["key", "value"])
lowercase_value_mapping = {"": None, "true": True, "false": False}


class EventTagFilter(EventFilter):
    def __init__(self, queries: list, min_level=EventLevel.INFO):
        self.queries = list()

        for query in queries:
            include = set()
            exclude = set()
            changes_include = set()
            changes_exclude = set()
            for m in re_match.finditer(query):
                token = m.group()
                if "=" in token:
                    changes = changes_include
                    if token[0] == "+":
                        token = token[1:]
                    elif token[0] == "-":
                        changes = changes_exclude
                        token = token[1:]

                    k, v = token.split("=", 1)
                    k = _unquote(k)
                    v = _unquote(v)
                    if not k:
                        raise AssertionError(
                            'Invalid filter query "%s", token: "%s"' % (query, token)
                        )

                    v = lowercase_value_mapping.get(v.lower(), v)

                    changes.add(KeyValue(k, v))
                elif token[0] == "+":
                    include.add(_unquote(token[1:]))
                elif token[0] == "-":
                    exclude.add(_unquote(token[1:]))
                else:
                    include.add(_unquote(token))

            self.queries.append(
                TagQuery(include, exclude, changes_include, changes_exclude)
            )

        super().__init__(min_level=min_level)

    def match(self, event: Event):
        tags = [event.type]

        if isinstance(event, LiveEvent):
            tags.append("live")
        elif isinstance(event, ChangeEvent):
            tags.append("change")

        if event.tags:
            tags += event.tags

        if isinstance(event, LiveEvent):
            tags += [event.name, event.label]

        return super().match(event) and any(
            all(i in tags for i in query.include)
            and all(e not in tags for e in query.exclude)
            and all(
                (
                    ci.key in event.change
                    and (ci.value is None or event.change.get(ci.key) == ci.value)
                )
                for ci in query.changes_include
            )
            and all(
                (
                    ce.key not in event.change
                    or (ce.value is not None and event.change.get(ce.key) != ce.value)
                )
                for ce in query.changes_exclude
            )
            for query in self.queries
        )


class LiveEventRegexpFilter(LiveEventFilter):
    def __init__(self, events_allow, events_ignore, min_level=EventLevel.INFO):
        self.events_allow = events_allow
        self.events_ignore = events_ignore

        super().__init__(min_level=min_level)

    def match(self, event: Event):
        allow = False
        if super().match(event):
            assert isinstance(event, LiveEvent)

            major_code = event.major
            minor_code = event.minor
            # Only let some elements pass

            for ev in self.events_allow:
                if isinstance(ev, tuple):
                    if major_code == ev[0] and (minor_code == ev[1] or ev[1] == -1):
                        allow = True
                        break
                elif isinstance(ev, str):
                    if re.match(ev, event.key):
                        allow = True
                        break

            # Ignore some events
            for ev in self.events_ignore:
                if isinstance(ev, tuple):
                    if major_code == ev[0] and (minor_code == ev[1] or ev[1] == -1):
                        allow = False
                        break
                elif isinstance(ev, str):
                    if re.match(ev, event.key):
                        allow = False
                        break

        return allow



import asyncio
import logging
from asyncio import Future
from collections import defaultdict
from typing import Awaitable, Callable, List, Mapping, Union

from paradox.event import Change, Event, Notification
from paradox.lib.utils import call_soon_in_main_loop

PREFIX = "pai_"

logger = logging.getLogger("PAI").getChild(__name__)
loop = asyncio.get_event_loop()


class Listener:
    def __init__(self, callback: Union[Callable, Future], **curriedArgs):
        if not isinstance(callback, (Callable, Future)):
            raise AssertionError("Wrong callback")
        self.callback = callback
        self.curriedArgs = curriedArgs

    async def call(self, **kwargs):
        kwargs2 = self.curriedArgs.copy()
        kwargs2.update(**kwargs)

        if isinstance(self.callback, Callable):
            result = self.callback(**kwargs2)
            if isinstance(result, Awaitable):
                await result
        elif isinstance(self.callback, Future):
            if not self.callback.done():
                self.callback.set_result(kwargs2)
            else:
                logger.debug("Future already done")
        else:
            raise Exception("Should not happen")

    def __eq__(self, other):
        if isinstance(other, Listener):
            return self.callback == other.callback
        return False


class PubSub:
    listeners: Mapping[str, List[Listener]]

    def __init__(self):
        self.listeners = defaultdict(list)

    def subscribe(self, listener: Callable, topicName: str, **curriedArgs):
        self.listeners[topicName].append(Listener(listener, **curriedArgs))

    def unsubscribe(self, listener: Callable, topicName: str):
        self.listeners[topicName].remove(Listener(listener))

    async def sendMessage(self, topicName: str, **msgData):
        return await asyncio.gather(
            *(l.call(**msgData) for l in self.listeners[topicName])
        )


pub = PubSub()


def subscribe(listener, topicName: str, **curriedArgs):
    pub.subscribe(listener, PREFIX + topicName, **curriedArgs)


def sendMessage(topicName: str, **msgData):
    call_soon_in_main_loop(pub.sendMessage(PREFIX + topicName, **msgData))


def sendEvent(event: Event):
    call_soon_in_main_loop(pub.sendMessage(PREFIX + "events", event=event))


def sendChange(change: Change):
    call_soon_in_main_loop(pub.sendMessage(PREFIX + "changes", change=change))


def sendNotification(notification: Notification):
    call_soon_in_main_loop(
        pub.sendMessage(PREFIX + "notifications", notification=notification)
    )



import binascii
import inspect
import logging
import typing

from construct import ChecksumError, Construct, Container, MappingError

from paradox.config import config as cfg
from paradox.exceptions import AuthenticationFailed, StatusRequestException

from . import parsers
from ..panel import Panel as PanelBase
from .event import event_map
from .property import property_map

logger = logging.getLogger("PAI").getChild(__name__)

ZONE_ACTIONS = dict(
    bypass={
        "flags": {"bypassed": True},
        "operation": "set",
    },
    clear_bypass={
        "flags": {"bypassed": True},
        "operation": "clear",
    },
    clear_alarm_memory={
        "flags": {"generated_alarm": True},
        "operation": "clear",
    },
)


class Panel_EVOBase(PanelBase):
    event_map = event_map
    property_map = property_map
    max_eeprom_response_data_length = 64
    status_request_addresses = parsers.RAMDataParserMap.keys()

    def __init__(
        self, core, start_communication_response, variable_message_length=True
    ):
        super().__init__(core, variable_message_length)

        self._populate_settings(start_communication_response)

    def _populate_settings(self, start_communication_response):
        raw_data = (
            start_communication_response.fields.data
            + start_communication_response.checksum
        )
        parsed = parsers.InitializeCommunication.parse(raw_data)
        if cfg.LOGGING_DUMP_MESSAGES:
            logger.debug("Second parse of InitializeCommunication response: %s", parsed)
        self.settings = parsed.fields.value

    def get_message(self, name: str) -> Construct:
        try:
            clsmembers = dict(inspect.getmembers(parsers))
            if name in clsmembers:
                return clsmembers[name]
        except ResourceWarning:
            pass

        return super().get_message(name)

    async def dump_memory(self, file, memory_type):
        """
        Dumps EEPROM and RAM memory to files
        :return:
        """
        if memory_type == "ram":
            await self.dump_memory_to_file(file, range(0, 59), True)
        elif memory_type == "eeprom":
            await self.dump_memory_to_file(file, range(0, 0xFFFF, 64))
        else:
            raise AttributeError(f"Unknown memory type: {memory_type}")

    async def dump_memory_to_file(self, file, range_, ram=False):
        mem_type = "RAM" if ram else "EEPROM"
        logger.info("Dump " + mem_type)

        def expect(command, address):
            return (
                lambda m: m.fields.value.po.command == command
                and m.fields.value.address == address
            )

        packet_length = 64  # 64 is max
        for address in range_:
            args = dict(
                address=address, length=packet_length, control=dict(ram_access=ram)
            )
            logger.info("Dumping %s: address %d" % (mem_type, address))
            reply = await self.core.send_wait(
                parsers.ReadEEPROM,
                args,
                reply_expected=expect(0x5, address),
            )

            if reply is None:
                logger.error("Could not read %s: address %d" % (mem_type, address))
                return

            data = reply.fields.value.data

            file.write(data)

    def parse_message(
        self, message: bytes, direction="topanel"
    ) -> typing.Optional[Container]:
        try:
            if message is None or len(message) == 0:
                return None

            parent_parsed = super().parse_message(message, direction)
            if parent_parsed:
                return parent_parsed

            if direction == "topanel":
                if message[0] == 0x70:
                    return parsers.CloseConnection.parse(message)
                elif message[0] == 0x00:
                    return parsers.InitializeCommunication.parse(message)
                elif message[0] == 0x30:
                    return parsers.SetTimeDate.parse(message)
                elif message[0] == 0x40:
                    return parsers.PerformPartitionAction.parse(message)
                elif message[0] == 0xD0:
                    return parsers.PerformZoneAction.parse(message)
            else:
                if message[0] >> 4 == 0x7:
                    return parsers.ErrorMessage.parse(message)
                elif message[0] >> 4 == 0x1:
                    return parsers.LoginConfirmationResponse.parse(message)
                elif message[0] >> 4 == 0x3:
                    return parsers.SetTimeDateResponse.parse(message)
                elif message[0] >> 4 == 4:  # Used for partitions and PGMs
                    return parsers.PerformActionResponse.parse(message)
                elif message[0] >> 4 == 0xD:
                    return parsers.PerformZoneActionResponse.parse(message)
                # elif message[0] == 0x50 and message[2] == 0x80:
                #     return PanelStatus.parse(message)
                # elif message[0] == 0x50 and message[2] < 0x80:
                #     return ReadEEPROM.parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] == 0x80:
                #     return PanelStatusResponse[message[3]].parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] < 0x80:
                elif message[0] >> 4 == 0x5:
                    return parsers.ReadEEPROMResponse.parse(message)
                # elif message[0] == 0x60 and message[2] < 0x80:
                #     return WriteEEPROM.parse(message)
                # elif message[0] >> 4 == 0x06 and message[2] < 0x80:
                #     return WriteEEPROMResponse.parse(message)
                elif message[0] >> 4 == 0xE:
                    if message[1] == 0xFF:
                        return parsers.LiveEvent.parse(message)
                    else:
                        return parsers.RequestedEvent.parse(message)

        except ChecksumError as e:
            logger.error("ChecksumError %s, message: %s", e, binascii.hexlify(message))
        except Exception:
            logger.exception("Exception parsing message: %s", binascii.hexlify(message))

        return None

    async def initialize_communication(self, password) -> bool:
        encoded_password = self.encode_password(password)

        self.settings.pc_password = encoded_password
        payload = parsers.InitializeCommunication.build(
            dict(fields=dict(value=self.settings))
        )

        logger.info("Installer login")
        reply = await self.core.send_wait(message=payload, reply_expected=[0x1, 0x0])

        if reply is None:
            logger.error("Installer login failed")
            return False

        if reply.fields.value.po.command == 0x0:
            logger.error(
                "Authentication Failed. Wrong PASSWORD. Make sure you use correct PC Password. In Babyware: "
                "Right click on your panel -> Properties -> PC Communication (Babyware) -> PC Communication "
                "(Babyware) Tab."
            )
            raise AuthenticationFailed("Wrong PASSWORD")

        if reply.fields.value.po.command == 0x1:
            logger.info("Authentication Success")

            if not reply.fields.value.po.status.Winload_connected:
                logger.warning("Winload_connected is still False")

            return True

        logger.error("Invalid response to InitializeCommunication: %s", reply)
        return False

    @staticmethod
    def _request_status_reply_check(message: Container, address: int):
        mvars = message.fields.value

        if (
            mvars.po.command == 0x5
            and mvars.control.ram_access is True
            and mvars.bus_address == 0x00  # panel
            and mvars.address == address
        ):
            return True

        return False

    async def request_status(self, i: int) -> typing.Optional[Container]:
        args = dict(address=i, length=64, control=dict(ram_access=True))
        reply = await self.core.send_wait(
            parsers.ReadEEPROM,
            args,
            reply_expected=lambda m: self._request_status_reply_check(
                m, args["address"]
            ),
        )
        if reply is not None:
            logger.debug("Received status response: %d" % i)
            return self.handle_status(reply, parsers.RAMDataParserMap)
        else:
            raise StatusRequestException("No reply to status request: %d" % i)

    async def control_partitions(self, partitions: list, command: str) -> bool:
        """
        Control Partitions
        :param list partitions: a list of partitions
        :param str command: textual command
        :return: True if we have at least one success
        """
        args = dict(partitions={i: command for i in partitions})

        try:
            reply = await self.core.send_wait(
                parsers.PerformPartitionAction, args, reply_expected=0x4
            )
        except MappingError:
            logger.error('Partition command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Partition command: "%s" succeeded' % command)
        else:
            logger.info('Partition command: "%s" failed' % command)
        return reply is not None

    async def control_zones(self, zones: list, command: str) -> bool:
        """
        Control zones
        :param list zones: a list of zones
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in ZONE_ACTIONS:
            return False

        args = ZONE_ACTIONS[command].copy()
        args["zones"] = zones

        try:
            reply = await self.core.send_wait(
                parsers.PerformZoneAction, args, reply_expected=0xD
            )
        except MappingError:
            logger.error('Zone command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Zone command: "%s" succeeded' % command)
        else:
            logger.info('Zone command: "%s" failed' % command)
        return reply is not None

    async def control_outputs(self, outputs, command) -> bool:
        """
        Control PGM
        :param list outputs: a list of pgms
        :param str command: textual command
        :return: True if we have at least one success
        """

        args = {"pgms": outputs, "command": command}

        try:
            reply = await self.core.send_wait(
                parsers.PerformPGMAction, args, reply_expected=0x4
            )
        except MappingError:
            logger.error('PGM command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('PGM command: "%s" succeeded' % command)
        else:
            logger.info('PGM command: "%s" failed' % command)
        return reply is not None

    async def control_doors(self, doors, command) -> bool:
        """
        Control Doors
        :param list doors: a list of doors
        :param str command: textual command
        :return: True if we have at least one success
        """

        args = {"doors": doors, "command": command}

        try:
            reply = await self.core.send_wait(
                parsers.PerformDoorAction, args, reply_expected=0x4
            )
        except MappingError:
            logger.error('Door command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Door command: "%s" succeeded' % command)
        else:
            logger.info('Door command: "%s" failed' % command)
        return reply is not None

    async def send_panic(self, partitions, panic_type, user_id):
        accepted = False

        args = {"partitions": partitions, "panic_type": panic_type, "user_id": user_id}

        reply = await self.core.send_wait(
            parsers.SendPanicAction, args, reply_expected=0x4
        )

        if reply is not None:
            accepted = True

        return accepted



from typing import Sequence, Union

from paradox.lib.utils import deep_merge


class ElementTypeContainer(dict):
    def __init__(self, *args, **kwargs):
        super(ElementTypeContainer, self).__init__(*args, **kwargs)

        self.key_index = {}

        self.reindex()

    def reindex(self):
        """Update additional text key index"""
        for key, value in self.items():
            if isinstance(value, dict) and "key" in value:
                self.key_index[value["key"]] = value

    def filter(self, keys: Sequence[Union[int, str]]):
        """
        Drops all elements that are not in `id_arr`. Does not support textual keys.
        :param Sequence[int] id_arr: Array of ints
        :return:
        """
        id_arr = self.select(keys)
        remove_keys = set(self.keys()) - set(id_arr)
        for i in remove_keys:
            try:
                self.__delitem__(i)
            except KeyError:
                pass

    def select(
        self, needle: Union[str, int, Sequence[Union[int, str]]]
    ) -> Sequence[int]:
        """
        Helper function to select objects from provided dictionary

        :param needle:
        :return: Sequence[int] list of object indexes
        """

        if needle == "all" or needle == "0":
            return list(self)

        else:
            if isinstance(needle, range):
                needle = list(needle)
            elif not isinstance(needle, (list, set)):
                needle = [needle]

            selected = []

            for n in needle:
                index = self.get_index(n)
                if index is not None:
                    selected.append(index)

            return selected

    def deep_merge(self, *dicts):
        deep_merge(self, *dicts)

    def get_index(self, key: Union[str, int]):
        if isinstance(key, int):
            return key
        e = self.get(key)
        if e is not None:
            for k, v in self.items():
                if v == e:
                    return k

    def get(self, key: Union[str, int], default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __contains__(self, key: Union[str, int]):
        item = self.get(key)
        return item is not None

    def __getitem__(self, key: Union[str, int]):
        key = self.__keytransform__(key)
        if isinstance(key, str):
            e = self.key_index.get(key)
            if e is not None:
                return e

        return super(ElementTypeContainer, self).__getitem__(key)

    def __setitem__(self, key: Union[str, int], value):
        key = self.__keytransform__(key)
        super(ElementTypeContainer, self).__setitem__(key, value)

        if isinstance(value, dict) and "key" in value:
            self.key_index[value["key"]] = value

    def __delitem__(self, key: Union[str, int]):
        key = self.get_index(key)
        try:
            el = self[key]
            if isinstance(el, dict) and "key" in el:
                text_key = el["key"]

                del self.key_index[text_key]
        except KeyError:
            pass

        super(ElementTypeContainer, self).__delitem__(self.__keytransform__(key))

    @staticmethod
    def __keytransform__(key: Union[str, int]):
        if isinstance(key, str) and key.isdigit():
            return int(key)
        return key



from ..panel import Panel_EVOBase


class Panel_EVO48(Panel_EVOBase):
    mem_map = {
        "labels": {
            "zone": {
                "label_offset": 0,
                "addresses": [range(0x00430, 0x00730, 0x10)],  # EVO48
            },
            "pgm": {
                "label_offset": 0,
                "addresses": [range(0x07082, 0x7482, 0x20)],
                "template": {"on": False, "pulse": False},
            },
            "partition": {
                "label_offset": 0,
                "addresses": [range(0x03A6B, 0x03C17, 0x6B)],  # EVO48
            },
            "user": {"label_offset": 0, "addresses": [range(0x03E47, 0x04447, 0x10)]},
            "module": {  # modules
                "label_offset": 0,
                "addresses": [range(0x04E47, 0x05637, 0x10)],  # EVO48
            },
            "door": {"label_offset": 0, "addresses": [range(0x0345C, 0x365C, 0x10)]},
        },
        "definitions": {
            "zone": {"addresses": [range(0x01F0, 0x024E + 2, 2)]},  # EVO48
            "partition": {
                "bit_encoded": True,
                "addresses": [[0x39D8]],  # All 8 partitions as bits
            },
            "user": {"addresses": [range(0x0BF0, 0x0FA6 + 10, 10),]},  # 96 users
        },
    }



import logging
from abc import abstractmethod

from paradox.connections.handler import ConnectionHandler
from paradox.connections.protocols import ConnectionProtocol
from paradox.lib.async_message_manager import AsyncMessageManager

logger = logging.getLogger("PAI").getChild(__name__)


class Connection(AsyncMessageManager, ConnectionHandler):
    _protocol: ConnectionProtocol

    def __init__(self):
        super().__init__()
        self.connected = False
        self._protocol: ConnectionProtocol = None

    def on_message(self, raw: bytes):
        self.schedule_raw_message_handling(raw)

    def on_connection(self):
        logger.info("Connection established")

    def on_connection_loss(self):
        logger.error("Connection was lost")
        self.connected = False

    @property
    def connected(self) -> bool:
        return (
            self._connected
            and self._protocol is not None
            and self._protocol.is_active()
        )

    @connected.setter
    def connected(self, value: bool):
        self._connected = value

    @abstractmethod
    async def connect(self) -> bool:
        raise NotImplementedError("Implement in a subclass")

    def write(self, data: bytes):
        if self.connected:
            self._protocol.send_message(data)  # throws ConnectionError
        else:
            raise ConnectionError("Not connected")

    async def close(self):
        if self._protocol:
            await self._protocol.close()
            self._protocol = None
        self.connected = False

    def variable_message_length(self, mode):
        if self._protocol is not None:
            self._protocol.variable_message_length(mode)



import logging
import os

import requests

from paradox.config import config as cfg
from paradox.event import EventLevel
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface

logger = logging.getLogger("PAI").getChild(__name__)


class HomeAssistantNotificationsTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Home Assistant"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.HOMEASSISTANT_NOTIFICATIONS_EVENT_FILTERS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_ALLOW_EVENTS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_IGNORE_EVENTS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_MIN_EVENT_LEVEL,
        )

        self.api_url = "http://supervisor/core/api/services/:domain/:service"
        if cfg.HOMEASSISTANT_NOTIFICATIONS_API_URL:
            self.api_url = cfg.HOMEASSISTANT_NOTIFICATIONS_API_URL

        self.token = os.environ.get("SUPERVISOR_TOKEN")
        if cfg.HOMEASSISTANT_NOTIFICATIONS_API_TOKEN:
            self.token = cfg.HOMEASSISTANT_NOTIFICATIONS_API_TOKEN
        if not self.token:
            logger.error(
                f'"SUPERVISOR_TOKEN" environment variable must be set to use {__class__.__name__}'
            )

    def send_message(self, message: str, level: EventLevel):
        if not self.token:
            logger.warning(
                "Unable to send a notification to Home Assistant. No token is set."
            )
            return

        url = self.api_url.replace(":domain", "notify").replace(
            ":service", cfg.HOMEASSISTANT_NOTIFICATIONS_NOTIFIER_NAME
        )

        data = {}

        if cfg.HOMEASSISTANT_NOTIFICATIONS_LOVELACE_URI:
            # iOS
            data["url"] = cfg.HOMEASSISTANT_NOTIFICATIONS_LOVELACE_URI
            # Android
            data["clickAction"] = cfg.HOMEASSISTANT_NOTIFICATIONS_LOVELACE_URI

        if level == EventLevel.CRITICAL:
            data.update(
                {
                    # iOS
                    "push": {
                        "interruption-level": "critical",
                    },
                    # Android
                    "ttl": 0,
                    "priority": "high",
                    "channel": "alarm_stream",
                }
            )

        payload = {"message": message, "title": "Paradox", "data": data}

        headers = {"Authorization": f"Bearer {self.token}"}

        res = requests.post(url, json=payload, headers=headers)
        if res.status_code == 200:
            logger.info(f"Notification sent: {message}, level={level}")
        else:
            logger.error(
                f"Failed to send notification: code={res.status_code}, text: {res.text}"
            )



import asyncio
import logging
from typing import Callable, Optional

from construct import Container

from paradox.config import config as cfg
from paradox.lib.handlers import FutureHandler, HandlerRegistry, PersistentHandler

logger = logging.getLogger("PAI").getChild(__name__)


class EventMessageHandler(PersistentHandler):
    def can_handle(self, data: Container) -> bool:
        assert isinstance(data, Container)
        values = data.fields.value
        return values.po.command == 0xE and (not hasattr(values, "requested_event_nr"))


class ErrorMessageHandler(PersistentHandler):
    def can_handle(self, data: Container) -> bool:
        assert isinstance(data, Container)
        return data.fields.value.po.command == 0x7 and hasattr(
            data.fields.value, "message"
        )


class AsyncMessageManager:
    def __init__(self):
        super().__init__()

        self.handler_registry = HandlerRegistry()
        self.raw_handler_registry = HandlerRegistry()

    async def wait_for_message(
        self,
        check_fn: Optional[Callable[[Container], bool]] = None,
        timeout=cfg.IO_TIMEOUT,
    ) -> Container:
        return await self.handler_registry.wait_until_complete(
            FutureHandler(check_fn), timeout
        )

    def register_raw_handler(self, handler):
        self.raw_handler_registry.append(handler)

    def deregister_raw_handler(self, name):
        self.raw_handler_registry.remove_by_name(name)

    def register_handler(self, handler):
        self.handler_registry.append(handler)

    def deregister_handler(self, name):
        self.handler_registry.remove_by_name(name)

    def schedule_message_handling(self, message: Container):
        return asyncio.get_event_loop().create_task(
            self.handler_registry.handle(message)
        )

    def schedule_raw_message_handling(self, message: Container):
        return asyncio.get_event_loop().create_task(
            self.raw_handler_registry.handle(message)
        )



import http.client
import logging
import re
import urllib

from paradox.config import config as cfg
from paradox.event import EventLevel
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface

logger = logging.getLogger("PAI").getChild(__name__)

_level_2_priority = {
    EventLevel.NOTSET: -2,
    EventLevel.DEBUG: -2,
    EventLevel.INFO: -1,
    EventLevel.WARN: 0,
    EventLevel.ERROR: 1,
    EventLevel.CRITICAL: 2,
}


class PushoverTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Pushover"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.PUSHOVER_EVENT_FILTERS,
            cfg.PUSHOVER_ALLOW_EVENTS,
            cfg.PUSHOVER_IGNORE_EVENTS,
            cfg.PUSHOVER_MIN_EVENT_LEVEL,
        )

        self.users = {}

    def send_message(self, message: str, level: EventLevel):
        for settings in cfg.PUSHOVER_BROADCAST_KEYS:
            user_key = settings["user_key"]
            devices_raw = settings["devices"]

            if devices_raw == "*" or devices_raw is None:
                self._send_pushover_message(user_key, message, level)
            else:
                devices = list(filter(bool, re.split(r"[\s]*,[\s]*", devices_raw)))

                for device in devices:
                    self._send_pushover_message(user_key, message, level, device)

    def _send_pushover_message(self, user_key, message, level, device=None):
        conn = http.client.HTTPSConnection("api.pushover.net:443")
        params = {
            "token": cfg.PUSHOVER_KEY,
            "user": user_key,
            "message": message,
            "priority": _level_2_priority.get(level, 0),
            "title": "Alarm",
        }
        if device:
            params["device"] = device

        conn.request(
            "POST",
            "/1/messages.json",
            urllib.parse.urlencode(params),
            {"Content-type": "application/x-www-form-urlencoded"},
        )

        response = conn.getresponse()
        if response.status != 200:
            logger.error(f"Failed to send message: {response.reason}")
        else:
            logger.info(
                f"Notification sent: {message}, level={level}, device={device if device else 'all'}"
            )
        conn.close()
